#include <complex>
#include <float.h>
#include <fstream>
#include <list>
#include <cmath>
#include <iostream>
#include "rmsd.h"
#include <stdlib.h>
#include <time.h>
#include "TNear.h"
#include <string.h>
#ifdef USE_SYSTEM_HEADERS
#include <vector_3d.h>
#else
#include "vector_3d.h"
#endif

const double pi = 4.0 * ::atan( 1.0 );
const double degreesPerRadians = pi / 180.0;
const double RadiansPerDegree = 180.0 / pi;


void TestOnePair( const Vector_3& x1, const Vector_3& x2, const Vector_3& z1, const Vector_3& z2, int& localErrorCount );
void TestAB_BestRotation( void );
void TestOne_ABBestRotation( void );
std::vector<Vector_3> BuildTruncatedIcosahedron( );
void TestKabsch( const std::vector<Vector_3> v1, const std::vector<Vector_3> v2 );

int ReadChain( std::vector<Vector_3>& v );

const Vector_3 xaxis( 1.0, 0.0, 0.0 );
const Vector_3 yaxis( 0.0, 1.0, 0.0 );
const Vector_3 zaxis( 0.0, 0.0, 1.0 );

long g_errorCount;

int main (int argc, char ** argv)
{
   g_errorCount = 0;
    TestAB_BestRotation( );

    if( g_errorCount == 0 )
    {
        fprintf(stdout,  "No errors were detected while testing Vector_3\n" );
    }

    return( g_errorCount );
}

int ReadChain( const char* filename, std::vector<long>& pdbAtomSerial, std::vector<Vector_3>& v )
{
   char line[100];
   memset( line, 0, sizeof(line) );
   std::ifstream file;
   file.open( filename );

   if( ! file.is_open( ))
   {
     g_errorCount++;
     fprintf(stdout," TestVector_3d: ReadChain: Failed to open file %s\n",filename);
   }

   // read and discard 3 header lines
   for ( unsigned int i=0; i<3; ++i )
   {
      file.getline( line, sizeof(line), '\n' );
   }

   double x,y,z;
   std::string junk;
   long serial;
   while ( file>>junk, file ) // discard the first item ("ATOM") and check that we aren't at eof
   {
      file>>serial;
      file>>x;
      file>>y;
      file>>z;
      pdbAtomSerial.push_back( serial );
      v.push_back( Vector_3(x,y,z) );
   }

   file.close( );

   return( v.size() );
}

void TestAB_BestRotation( void )
{
   std::vector<Vector_3>  vlist1;
   std::vector<Vector_3>  vlist2;
   std::vector<long> pdbAtomSerial1;
   std::vector<long> pdbAtomSerial2;
   Vector_3 cm1,cm2;
   double rmsd;
   Matrix_3x3 bestRotation;

   //{
   //   // treat one point
   //   vlist1.clear( );
   //   vlist2.clear( );
   //   vlist1.push_back( Vector_3(1,0,0) );
   //   vlist2.push_back( Vector_3(0,1,0) );
   //   const std::vector<double> quaternion = AB_BestRotation( vlist1, vlist2, cm1, cm2, bestRotation, rmsd, 0 );
   //}
   //{
   //   // treat two points
   //   vlist1.clear( );
   //   vlist2.clear( );
   //   vlist1.push_back( Vector_3(1,0,0) );
   //   vlist1.push_back( Vector_3(2,0,0) );

   //   vlist2.push_back( Vector_3(0,1,0) );
   //   vlist2.push_back( Vector_3(0,2,0) );
   //   const std::vector<double> quaternion = AB_BestRotation( vlist1, vlist2, cm1, cm2, bestRotation, rmsd, 0 );
   //}
   //{
   //   // treat three points
   //   vlist1.clear( );
   //   vlist2.clear( );
   //   vlist1.push_back( Vector_3(1,0,0) );
   //   vlist1.push_back( Vector_3(2,0,0) );
   //   vlist1.push_back( Vector_3(3,0,0) );

   //   vlist2.push_back( Vector_3(0,1,0) );
   //   vlist2.push_back( Vector_3(0,2,0) );
   //   vlist2.push_back( Vector_3(0,3,0) );
   //   const std::vector<double> quaternion = AB_BestRotation( vlist1, vlist2, cm1, cm2, bestRotation, rmsd, 0 );
   //}
   {
      vlist1.clear( );
      vlist2.clear( );
      //const int chainACount = ReadChain( "iknvA-coord-CA.pdb", vlist1 );
      //const int chainBCount = ReadChain( "iknvB-coord-CA.pdb", vlist2 );
      const int chainACount = ReadChain( "1h9mA.pdb", pdbAtomSerial1, vlist1 );
      const int chainBCount = ReadChain( "1h9mB.pdb", pdbAtomSerial2, vlist2 );
      int nvec;
      if ( chainACount != chainBCount )
      {
         nvec = 0;
      }
      else
      {
         nvec = chainACount;
      }

      // get the answer from Bosco Ho's rmsd code
    TestKabsch( vlist1, vlist2 );

    {
        // use AB_BestRotation
          Vector_3 cm1, cm2;
          Matrix_3x3 m;
          double rmsd;
          QR<double> result = AB_BestRotation( vlist1, vlist2, cm1, cm2, m, rmsd, 0 );

    }

   //   
   //const Matrix_3x3 m = Vector_3( 0.0, 0.0, 1.0 ).Rotmat( 90.0/180.0/(4.0*atan(1.0)) );

   //for ( int i=0; i<chainACount; ++i )
   //{
   //   vlist2[i] = m.MV( vlist1[i] );
   //}

      //const std::vector<double> quaternion = AB_BestRotation( vlist1, vlist2, cm1, cm2, bestRotation, rmsd, 0 );

      // CASES
      // pairCount > 0 => ignore singleAtomSelect
      // pairCount == 0 and singleAtomSelect < 0 => exhaustive search
      // pairCount == 0 and singleAtomSelect >=0 => do all compares to the selected atom and no more
      CNearTree<Vector_3> tree;
      const int pairCount = 33;
      const int singleAtomSelect = 299;
      std::vector<Vector_3> centered_v1, centeredRotated_v2, axisVectorList, axisBaseVectorList;
      std::vector<QR<double> > quaternionList;
      std::vector<std::pair< std::pair<long,long>, std::pair<long,long> > > pdbSerials;
      std::vector<std::pair< long, long > > indices;
      const QR<double> quaternion = AB_BestRotationCore( 
         vlist1, 
         vlist2, 
         cm1, 
         cm2, 
         pdbAtomSerial1,
         pdbAtomSerial2,
         centered_v1,
         centeredRotated_v2,
         pdbSerials,
         indices,
         quaternionList,
         axisVectorList,
         axisBaseVectorList,
         bestRotation, 
         rmsd, 
         pairCount,
         singleAtomSelect
         );

      fprintf( stdout, "There were %d atoms in each input file\n", pdbAtomSerial1.size( ) );
      fprintf( stdout, "There were %d interactions calculated\n\n", axisVectorList.size( ) );
      fprintf( stdout, "rmsd %f \n", rmsd );

      const Vector_3 vBestFromQuaternion( quaternion[1], quaternion[2], quaternion[3] );

      tree.clear( );
      tree.insert( BuildTruncatedIcosahedron( ) );
      tree.CompleteDelayedInsertRandom( );
      std::vector<long> sphereGridCounts( tree.size( ), 0 );

      // find the actual length of a side
      std::vector<Vector_3> vFound;
      const unsigned long nFound = tree.FindInAnnulus( .001, 0.5, vFound, tree[0] );
      double minRadius = DBL_MAX;
      for( unsigned int i=0; i<nFound-1; ++i )
      {
          for ( unsigned int j=i+1; j<nFound; ++j )
          {
              minRadius = std::min( minRadius, (vFound[i]-vFound[j]).Norm( ) );
              const int i19191 = 19191;
          }
      }

      for ( unsigned int i=0; i<axisVectorList.size(); ++i )
      {
          CNearTree<Vector_3>::iterator it = tree.NearestNeighbor( minRadius, axisVectorList[i] );
          if( it != tree.end( ) )
          {
              ++sphereGridCounts[it.get_position( )];
          }
      }

      int maxValue = INT_MIN;
      unsigned int maxIndex = 0;
      for( unsigned int i=0; i<sphereGridCounts.size( ); ++i )
      {
          if ( maxValue < sphereGridCounts[i] )
          {
              maxValue = sphereGridCounts[i];
              maxIndex = i;
          }
          fprintf( stdout, "%d \t\t%f %f %f\n", sphereGridCounts[i], tree[i][0], tree[i][1], tree[i][2] );
      }

      CNearTree<Vector_3> axisTree( axisVectorList );;
      std::vector<Vector_3> nearBestPoint;
      axisTree.FindInSphere( 0.9*minRadius, nearBestPoint, tree[maxIndex] );
      const Vector_3 vBestFromAxes = Vector_3::GetCenterOfMass( nearBestPoint );
      fprintf( stdout, "vBestFromQuaternion %f %f %f \n", vBestFromQuaternion[0], vBestFromQuaternion[1], vBestFromQuaternion[2] );
      fprintf( stdout, "vBestFromAxes %f %f %f \n", vBestFromAxes[0], vBestFromAxes[1], vBestFromAxes[2] );


      for ( unsigned int i=0; i<indices.size(); ++i )
      {
          const std::pair<long,long> p1 = pdbSerials[i].first;
          const std::pair<long,long> p2 = pdbSerials[i].second;
          fprintf( stdout, " %d %d %d %d %f %f  %f %f %f  %f %f %f\n", 
              p1.first, p1.second,
              p2.first, p2.second,
              atan2(sqrt(1-quaternionList[i][0]*quaternionList[i][0]) ,quaternionList[i][0])*180.0/4.0/atan(1.0),
              (vlist1[indices[i].first]-vlist1[indices[i].second]).Norm(),
              axisVectorList[i][0], axisVectorList[i][1], axisVectorList[i][2],
              axisBaseVectorList[i][0], axisBaseVectorList[i][1], axisBaseVectorList[i][2]);
      }
   }
}

void TestKabsch( const std::vector<Vector_3> v1, const std::vector<Vector_3> v2 )
{
    double x1[1000][3];
    double x2[1000][3];
    for ( unsigned int i=0; i<v1.size( ); ++i )
    {
        x1[i][0] = v1[i][0];
        x1[i][1] = v1[i][1];
        x1[i][2] = v1[i][2];
        x2[i][0] = v2[i][0];
        x2[i][1] = v2[i][1];
        x2[i][2] = v2[i][2];
    }
    double mov[3], ref[3], U[3][3];
    double rmsd;
    calculate_rotation_rmsd( x1,x2, v1.size( ), mov, ref, U, &rmsd );

    fprintf( stdout, "rmsd from Kabsch %f\n", rmsd );
}

void TestOne_ABBestRotation( void )
{
   //std::vector<Vector_3>  vlist1;
   //const int chainACount = ReadChain( "1h9mA.pdb", vlist1 );

   //std::vector<Vector_3>  vlist2;
   //const int chainBCount = ReadChain( "1h9mB.pdb", vlist2 );

   //const Matrix_3x3 m = Vector_3( 0.0, 0.0, 1.0 ).Rotmat( 90.0/180.0/(4.0*atan(1.0)) );

   //for ( int i=0; i<chainACount; ++i )
   //{
   //   vlist2[i] = m.MV( vlist1[i] );
   //}

   //int nvec;
   //if ( chainACount != chainBCount )
   //{
   //   nvec = 0;
   //}
   //else
   //{
   //   nvec = chainACount;
   //}

   //Vector_3 cm1,cm2;
   //double rmsd;
   //Matrix_3x3 bestRotation;
   //
   //const std::vector<double> quaternion = AB_BestRotation( vlist1, vlist2, cm1, cm2, bestRotation, rmsd, 0 );
   //
   // // test Kabsch's algorithm
   //double mov_com[3];
   //double mov_to_ref[3];
   //double U[3][3];
   //double rmsd_Kabsch;

   //double v1[1000][3];
   //double v2[1000][3];
   //for ( int i=0; i<nvec; ++i )
   //{
   //   v1[i][0] = vlist1[i][0];
   //   v1[i][1] = vlist1[i][1];
   //   v1[i][2] = vlist1[i][2];

   //   v2[i][0] = vlist2[i][0];
   //   v2[i][1] = vlist2[i][1];
   //   v2[i][2] = vlist2[i][2];
   //}

   // calculate_rotation_rmsd( v1, v2, nvec, mov_com, mov_to_ref, U, &rmsd_Kabsch );

   //const Matrix_3x3 mkabsch( U[0][0],U[0][1],U[0][2], U[1][0],U[1][1],U[1][2], U[2][0],U[2][1],U[2][2] );
   //double sumk = 0.0;

   //for( int i=0; i<nvec; ++i )
   //{
   //   sumk += (vlist1[i]-vlist2[i].MV(mkabsch)).SquaredLength();
   //}
   //sumk = sqrt(sumk/nvec);

}

std::vector<Vector_3> BuildTruncatedIcosahedron( )
{
   //Cartesian coordinates for the vertices of a truncated icosahedron 
   //centered at the origin are the orthogonal rectangles (0,1,3?), 
   //(1,3?,0), (3?,0,1) and the orthogonal cuboids (2,(1+2?),?), 
   //((1+2?),?,2), (?,2,(1+2?)) along with the orthogonal 
   //cuboids (1,(2+?),2?), ((2+?),2?,1), (2?,1,(2+?)), 
   //where ? = (1+?5)/2 is the golden mean. Using ?2 = ? + 1 one 
   //verifies that all vertices are on a sphere, centered at the 
   //origin, with the radius squared equal to 9? + 10. The 
   //edges have length 2.

   const double t = (1.0+sqrt(5.0))/2.0;

   std::vector<Vector_3> v;

   {
      v.push_back( Vector_3( 0,  1,  3.0*t ) ); // (0,1,3?)
      v.push_back( Vector_3( 0, -1,  3.0*t ) );
      v.push_back( Vector_3( 0,  1, -3.0*t ) );
      v.push_back( Vector_3( 0, -1, -3.0*t ) );

      v.push_back( Vector_3(  1,  3.0*t, 0 ) ); // (1,3?,0)
      v.push_back( Vector_3( -1,  3.0*t, 0 ) );
      v.push_back( Vector_3(  1, -3.0*t, 0 ) );
      v.push_back( Vector_3( -1, -3.0*t, 0 ) );

      v.push_back( Vector_3(  3.0*t, 0,  1 ) ); // (3?,0,1)
      v.push_back( Vector_3( -3.0*t, 0,  1 ) );
      v.push_back( Vector_3(  3.0*t, 0, -1 ) );
      v.push_back( Vector_3( -3.0*t, 0, -1 ) );
   }

   {
      v.push_back( Vector_3(  2, (1+2*t),  t ) ); // (2,(1+2?),?)
      v.push_back( Vector_3( -2, (1+2*t),  t ) );
      v.push_back( Vector_3(  2,-(1+2*t),  t ) );
      v.push_back( Vector_3(  2, (1+2*t), -t ) );
      v.push_back( Vector_3( -2,-(1+2*t),  t ) );
      v.push_back( Vector_3( -2, (1+2*t), -t ) );
      v.push_back( Vector_3(  2,-(1+2*t), -t ) );
      v.push_back( Vector_3( -2,-(1+2*t), -t ) );
   }
   
   {
      v.push_back( Vector_3(  (1+2*t),  t,  2 ) ); // ((1+2?),?,2)
      v.push_back( Vector_3( -(1+2*t),  t,  2 ) ); //
      v.push_back( Vector_3(  (1+2*t), -t,  2 ) ); //
      v.push_back( Vector_3(  (1+2*t),  t, -2 ) ); //
      v.push_back( Vector_3( -(1+2*t), -t,  2 ) ); // 
      v.push_back( Vector_3( -(1+2*t),  t, -2 ) ); // 
      v.push_back( Vector_3(  (1+2*t), -t, -2 ) ); // 
      v.push_back( Vector_3( -(1+2*t), -t, -2 ) ); //
   }

   {
      v.push_back( Vector_3(  t,   2,   (1+2*t) ) ); // (?,2,(1+2?))
      v.push_back( Vector_3( -t,   2,   (1+2*t) ) ); 
      v.push_back( Vector_3(  t,  -2,   (1+2*t) ) ); 
      v.push_back( Vector_3(  t,   2,  -(1+2*t) ) ); 
      v.push_back( Vector_3( -t,  -2,   (1+2*t) ) ); 
      v.push_back( Vector_3( -t,   2,  -(1+2*t) ) ); 
      v.push_back( Vector_3(  t,  -2,  -(1+2*t) ) ); 
      v.push_back( Vector_3( -t,  -2,  -(1+2*t) ) ); 
   }
   {
      v.push_back( Vector_3(  1,  (2+t),  2*t ) ); // (1,(2+?),2?)
      v.push_back( Vector_3( -1,  (2+t),  2*t ) ); 
      v.push_back( Vector_3(  1, -(2+t),  2*t ) ); 
      v.push_back( Vector_3(  1,  (2+t), -2*t ) ); 
      v.push_back( Vector_3( -1, -(2+t),  2*t ) ); 
      v.push_back( Vector_3( -1,  (2+t), -2*t ) ); 
      v.push_back( Vector_3(  1, -(2+t), -2*t ) ); 
      v.push_back( Vector_3( -1, -(2+t), -2*t ) ); 
   }

   {
      v.push_back( Vector_3(  (2+t),  2*t,  1 ) ); // ((2+?),2?,1) 
      v.push_back( Vector_3( -(2+t),  2*t,  1 ) ); 
      v.push_back( Vector_3(  (2+t), -2*t,  1 ) ); 
      v.push_back( Vector_3(  (2+t),  2*t, -1 ) ); 
      v.push_back( Vector_3( -(2+t), -2*t,  1 ) ); 
      v.push_back( Vector_3( -(2+t),  2*t, -1 ) ); 
      v.push_back( Vector_3(  (2+t), -2*t, -1 ) ); 
      v.push_back( Vector_3( -(2+t), -2*t, -1 ) ); 
   }

   {
      v.push_back( Vector_3(  2*t,  1,  (2+t) ) ); // (2?,1,(2+?)), 
      v.push_back( Vector_3( -2*t,  1,  (2+t) ) );
      v.push_back( Vector_3(  2*t, -1,  (2+t) ) );
      v.push_back( Vector_3(  2*t,  1, -(2+t) ) );
      v.push_back( Vector_3( -2*t, -1,  (2+t) ) );
      v.push_back( Vector_3( -2*t,  1, -(2+t) ) );
      v.push_back( Vector_3(  2*t, -1, -(2+t) ) );
      v.push_back( Vector_3( -2*t, -1, -(2+t) ) );
   }

   int countSideEquals2 = 0;
   for ( unsigned int i=0; i<v.size()-1; ++i )
   {
      for ( unsigned int j=i+1; j<v.size( ); ++j )
      {
         if( abs((v[i]-v[j]).Norm( )-2.0) < 0.001 ) ++countSideEquals2;
      }
   }

   if ( countSideEquals2 != 90 )
   {
      // some coordinate must be wrong or missing
   }

   const double radius = v[0].Norm( );
   // project everything onto the unit sphere
   for ( unsigned int i=0; i<v.size( ); ++i )
   {
      v[i] = v[i] / radius;
   }

   return( v );
}
