C23456789 123456789 123456789 123456789 123456789 123456789 123456789 12
Cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
C  VECTOR   -- Tools for 3-D Geometry
C
C this work may be freely distributed and incorporated within programs
C provided it is not for direct profit.  This message may NOT be removed
C nor may this work be copyrighted by other than the author(s).
C Lawrence C. Andrews   August 18, 1990    National Cancer Institute
C Frederick Cancer Research and Development Center  (sic)
C
C While the information contained in this document is correct to the 
C best of our knowledge,  we assume no responsibilities for any
C errors or inaccuracies.
Ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

C**********************************************************************C
      SUBROUTINE PAIRSM (V1,V2,AMAT,VMAT)
C----  SUBROUTINE PAIRSM AND PAIREN TO DO THE SUMS AND FINAL CALC-
C    ULATIONS FOR LEAST SQUARES FITTING
C
C  THE ALGORITHM IS FROM TICHY, 1970, ACTA CRYST V. A
C  IT HAS BEEN MODIFIED FROM A 9X9 LEAST SQUARES PROBLEM WITH BLOCK
C  FORM TO A PRODUCT OF TWO 3X3 MATRICES
C
C----THE RESULT OF THE USE OF THESE TWO SUBROUTINE WILL BE A MATRIX
C    WHICH GIVES THE BEST FIT OF THE VECTOR INPUT AS V1 TO THOSE INPUT
C    AS V2----THAT IS IT FINDS THE MATRIX WHICH WHEN MULTIPLIED TIMES
C    V1 GIVES V2 (IN THE LEAST SQUARES SENSE)
C
      DIMENSION V1(3),V2(3),AMAT(9),VMAT(9)
      REAL VV(3)
C----------------------------------------------------------------------C
      J = 0
      DO 1000 I=1,9,3
      J = J + 1
      CALL CONMUL (V1,V2(J),VV)
      CALL VECSUM (VV,VMAT(I),VMAT(I))
      CALL CONMUL (V1,V1(J),VV)
      CALL VECSUM (VV,AMAT(I),AMAT(I))
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE PAIREN(AMAT,VMAT,BMAT)
C  FINISHES THE LEAST SQUARES PROBLEM BEGUN USING PAIRSM
C
C----AMAT IS THE MATRIX OF THE NORMAL EQUATIONS
C        (JUST ONE OF THE THREE IDENTICAL DIAGONAL BLOCKS)
C    VMAT IS THE VECTOR OF THE NORMAL EQUATIONS
C    BMAT IS THE MATRIX OF TRANSFORMATION FOR THE BEST L.S. FIT
      DIMENSION AMAT(9),VMAT(9),BMAT(9),CMAT(9),DMAT(9)
C----------------------------------------------------------------------C
      CALL INVER (AMAT,CMAT)
      CALL TRNSPZ (CMAT,DMAT)
      CALL MATMUL (VMAT,DMAT,BMAT)
      END

C**********************************************************************C
      SUBROUTINE PAIR (X1,X2,Y1,Y2,AMAT)
C  RETURNS AMAT AS THE MATRIX WHICH ROTATES X1 TO COINCIDE WITH Y1
C  AND PUT X2 IN THE PLANE OF Y1,Y2 AS CLOSE TO Y2 AS POSSIBLE
      DIMENSION X1(3),X2(3),Y1(3),Y2(3),V(3),AMAT(9)
     2  ,BMAT(9),CMAT(9),ZERO(3)
      DATA ZERO /0.0, 0.0, 0.0/
C----------------------------------------------------------------------C
      AMX1 = AMAG(X1)
      AMY1 = AMAG(Y1)
      IF (AMX1 .LT. 1.0E-20 .OR. AMY1 .LT. 1.0E-20) THEN
         DO 1000 I=1,9
            AMAT(I) = 0.0
 1000    CONTINUE
         DO 2000 I=1,9,4
            AMAT(I) = 1.0
 2000    CONTINUE
      ELSE
         CALL CROSS (X1,Y1,V)
         IF ((AMAG(V) / (AMAG(X1) * AMAG(Y1))) .LT. 1.0E-4)
     2    CALL CROSS (X2,Y1,V)
         D = -TORSIN (X1,V,ZERO,Y1) / 57.2957795
         CALL ROTMAT (V,D,BMAT)
         CALL VECMUL (X2,BMAT,V)
         D = -TORSIN (V,Y1,ZERO,Y2) / 57.2957795
         CALL ROTMAT (Y1,D,CMAT)
         CALL MATMUL (CMAT,BMAT,AMAT)
      ENDIF
      END

C**********************************************************************C
      FUNCTION INSYM (ALINE,N,AMAT,V)
C
C  TRANSLATES TEXT DESCRIPTIONS OF SYMMETRY ELEMENTS INTO MATRIX,
C  VECTOR FORM
C
C  PARSES A CHARACTER STRING ALINE OF N CHARACTERS AND RETURNS THE
C  MATRIX AND VECTOR CORRESPONDING TO THE SYMMETRY OPERATION.  IF
C  THE TRANSLATION IS SUCCESSFUL, THEN INSYM IS RETURNED AS 0
C  IF THE STRING IS NOT TRANSLATABLE, THEN 1 IS RETURNED.  IF THE
C  STRING IS CAPABLE OF BEING PARSED, BUT IT DOES NOT REPRESENT
C  A CORRECT SPACE GROUP OPERATION, THEN 2 IS RETURNED, AND
C  THE MATRIX AND VECTOR ARE ALSO RETURNED.


      CHARACTER ALINE(N), CHAR(13), BLANK
      DIMENSION AMAT(9),  V(3)
      DIMENSION TRANS(9)
      DIMENSION NUM(2 )
      DATA NCHAR /13/
      DATA CHAR /'1','2','3','4','5','6','X','Y','Z','/','+','-',',' /
      DATA BLANK / ' ' /
      DATA TRANS /.5, .3333333, .25, .1666667, .6666667, .75,.8333333 ,
     2 0.0, 1.0 /
C----------------------------------------------------------------------C
      INSYM = 0
      N1 = N + 1
      DO 100 I=1,9
  100 AMAT(I) = 0.0
      DO 200 I=1,3
  200 V(I) = 0.0
      ISTART = 1
      DO 4000 IROW = 1,9,3
      INUM = 1
      NUM(1) = 1
      NUM(2) = 1
      ICOLUM = 0
      DO 2000 I=ISTART,N1
C----IF THERE IS NO TERMINAL COMMA, QUIT ANYWAY
      IF (I .EQ. N1)  GO TO 1200
      IF (ALINE(I) .EQ. BLANK) GO TO 2000
      DO 1000 IC=1,NCHAR
      II = IC
      IF (ALINE(I) .EQ. CHAR(II)) then
         GO TO 1100
      ELSEIF (ALINE(I) .EQ. 'x') then
         II = 7
         GO TO 1100
      ELSEIF (ALINE(I) .EQ. 'y') then 
         II = 8
         GO TO 1100
      ELSEIF (ALINE(I) .EQ. 'z') then
         II = 9
         GO TO 1100
      ENDIF
 1000 CONTINUE
      WRITE (*,*) I,' ',ALINE(I)
      INSYM = 1
      GO TO 9999
 1100 IF (II .LE. 6) NUM(INUM) = II * NUM(INUM)
      IF (IABS(II-8) .LE. 1) ICOLUM = II - 6
      IF (II .EQ. 10) INUM = 2
      IF (II .LE. 10) GO TO 2000
 1200 IR = (IROW+3) / 3
      DECIM = FLOAT(NUM(1)) / FLOAT(NUM(2))
      IF (ICOLUM .NE. 0) AMAT(IROW+ICOLUM-1) = DECIM
      IF (NUM(1) .EQ. NUM(2)) DECIM = 0.0
      IF (ICOLUM .EQ. 0) V(IR) = DECIM
      IF (II .EQ. 13) GO TO 4000
      NUM(1) = 1
      NUM(2) = 1
      IF (II .EQ. 12) NUM(1) = -1
      ICOLUM = 0
      INUM = 1
 2000 CONTINUE
      GO TO 9000
 4000 ISTART = I + 1
 9000 D = DET(AMAT)
      IF (ABS(1.0 - ABS(D)) .GT. 0.001) INSYM = 2
      DO 9200 II=1,3
      DO 9100 I=1,9
      IF (ABS(ABS(AMOD(V(II),1.0))-TRANS(I)) .LT. 0.0001) GO TO 9200
 9100 CONTINUE
      INSYM = 2
 9200 CONTINUE
 9999 RETURN
      END

C**********************************************************************C
      SUBROUTINE INRTIA (Y,WT,TEN)
C----SUM THE ELEMENTS OF AN INERTIA TENSOR, ONE ELEMENT PER CALL
C----THE EIGENVECTOR OF THE LARGEST EIGENVALUE IS THE BEST PLANE
C----THE EIGENVECTOR OF THE LARGEST EIGENVALUE OF THE INVERSE MATRIX
C     IS THE BEST LINE (AND WORST PLANE)
      REAL X(3)
      DIMENSION Y(3),TEN(9)
C----------------------------------------------------------------------C
      J = 0
      W = SQRT(WT)
      CALL CONMUL (Y,W,X)
      AM = DOT(X,X)
      DO 1000 I=1,9,4
      J = J + 1
      TEN(I) = TEN(I) + AM - X(J) * X(J)
 1000 CONTINUE
      TEN(2) = TEN(2) - X(1) * X(2)
      TEN(4) = TEN(2)
      TEN(3) = TEN(3) - X(1) * X(3)
      TEN(7) = TEN(3)
      TEN(6) = TEN(6) - X(2) * X(3)
      TEN(8) = TEN(6)
      END

C**********************************************************************C
      FUNCTION EIGEN1(A,V,B)
C  THE ALGORITHM IS THAT OF ANDREWS AND BERNSTEIN, 1976, ACTA CRYST
C   A32, 504-506
      DIMENSION X(9)
      DIMENSION A(9), B(9), V(3)
      DIMENSION Y(9)
C----------------------------------------------------------------------C
      CALL CPYMAT (A,B)
      DO 4500 IEIGEN=1,18
      T = B(1) + B(5) + B(9)
      IF (ABS(T) .LT. 1.0E-8) T = 1.0
      DO 3000 I=1,9
      X(I) = B(I) / T
 3000 CONTINUE
      CALL CPYMAT(X,Y)
      CALL MATMUL(Y,X,B)
      TEST = 0.0
      DO 4000 I=1,9
      D = X(I) - B(I)
      TEST = TEST + D * D
 4000 CONTINUE
 4500 CONTINUE
C----NOW GET THE EIGENVALUE
      CALL MATMUL(A,X,B)
      EIGEN1 = 0.0
      N = 0
      DO 5000 I=1,9
      IF (ABS(X(I)) .LT. 1.0E-18) GO TO 5000
      N = N + 1
      EIGEN1 = EIGEN1 + B(I)/X(I)
 5000 CONTINUE
      IF (N .NE. 0) EIGEN1 = EIGEN1 / FLOAT(N)
C-----COPY THE DEFLATED MATRIX FOR RETURNING TO CALL
      CALL CPYMAT (X,B)
      N = 1
      IF (AMAG(X(4)) .GT. AMAG(X(N))) N = 4
      IF (AMAG(X(7)) .GT. AMAG(X(N))) N = 7
      CALL UNIT (X(N),V)
      END

C**********************************************************************C
      SUBROUTINE GENROT (V1,V2,ANG,AMAT,TRAN)
C FOR TWO POINTS V1 AND V2, RETURNS THE MATRIX AND THE VECTOR FOR A 
C ROTATION OF ANG RADIANS ABOUT THE LINE FROM V1 TO V2
C
C   TO ROTATE A VECTOR X. MULTIPLY IT BY AMAT, AND THEN ADD TRAN
      DIMENSION X(3),Y(3)
C----------------------------------------------------------------------C
      CALL VECDIF (V2,V1,X)
      CALL ROTMAT (X,ANG,AMAT)
      CALL VECMUL (V1,AMAT,Y)
      CALL VECDIF (V1,Y,TRAN)
      END

C**********************************************************************C
      SUBROUTINE UNITRY (A,B)
C----GRAM-SCHMIDT ORTHOGONALIZATION OF A RETURNED IN B
      DIMENSION A(9),B(9)
C----------------------------------------------------------------------C
      CALL CROSS (A(1),A(4),B(7))
      CALL CROSS (B(7),A(1),B(4))
      CALL UNIT (B(7),B(7))
      CALL UNIT (B(4),B(4))
      CALL UNIT (A(1),B(1))
      END

C**********************************************************************C
      SUBROUTINE ROTMAT (V,ANG,AMAT)
C----SUBROUTINE TO RETURN A ROTATION MATRIX GIVEN A VECTOR AND AN ANGLE
C     OF ROTATION (IN RADIANS)ABOUT THAT VECTOR
      DIMENSION X(3)
      DIMENSION V(3)
      DIMENSION AXIS (5)
      DIMENSION AMAT(9), BMAT(9)
      DATA AXIS /0.0, 0.0, 1.0, 0.0, 0.0 /
C----------------------------------------------------------------------C
      SINANG = SIN(ANG)
      COSANG = COS (ANG)
      J = 4
      CALL UNIT (V,X)
      DO 1000 I=1,9,3
      J = J - 1
      CALL SUBROT (X,AXIS(J),SINANG,COSANG,BMAT(I))
 1000 CONTINUE
      CALL TRNSPZ (BMAT,AMAT)
      END

C**********************************************************************C
      SUBROUTINE SUBROT (V,AXIS,SINANG,COSANG,ROW)
      DIMENSION V(3)
      DIMENSION X(3),Y(3)
C----------------------------------------------------------------------C
      CALL CROSS (V,AXIS,X)
      IF (AMAG(X) - 1.0E-8) 2000, 2000, 1000
 1000 CALL CROSS (X,V,Y)
      CALL CONMUL (X,SINANG,X)
      CALL CONMUL (Y,COSANG,Y)
      CALL VECSUM (X,Y,X)
      F = DOT(V,AXIS)
      CALL CONMUL (V,F,Y)
      CALL VECSUM (X,Y,ROW)
C----THIS CALL TO UNIT IS ONLY TO REMOVE ROUNDING ERRORS
      CALL UNIT (ROW,ROW)
      GO TO 3000
 2000 CALL CPYVEC (AXIS,ROW)
 3000 RETURN
      END

C**********************************************************************C
      SUBROUTINE SOLVE (A,B,C,V)
      DIMENSION A(4),B(4),C(4),V(3),AMAT(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,3
      CALL MAT (AMAT,A,B,C)
      DD = DET(AMAT)
      IF (ABS(DD) .LT. 1.E-4) GO TO 9000
      AMAT(I)  = A(4)
      AMAT(I+3) = B(4)
      AMAT(I+6) = C(4)
      V(I) = DET(AMAT)/DD
 1000 CONTINUE
      RETURN
 9000 DO 9100 I=1,3
      V(I) = 1000000.0
 9100 CONTINUE
      END

C**********************************************************************C
      FUNCTION DET(A)
C----RETURN THE VALUE OF THE DETERMINANT OF A MATRIX
      REAL X(3)
      DIMENSION A(9)
C----------------------------------------------------------------------C
      CALL CROSS (A(1),A(4),X)
      DET = DOT (X,A(7))
      END

C**********************************************************************C
      SUBROUTINE VECMUL (X,UB,H)
C----MULTIPLY A VECTOR TIMES A MATRIX AND RETURN
C      THE NEW VECTOR IN H
      DIMENSION UB(9)
      DIMENSION H(3),X(3)
C----------------------------------------------------------------------C
      J = 0
      DO 1000 I=1,9,3
      J = J + 1
      H(J) =  DOT (UB(I),X)
 1000 CONTINUE
      END

C**********************************************************************C
      FUNCTION ALNGTH (A,B)
C-----GET THE SEPARATION BETWEEN THE ENDS OF TWO VECTORS
      REAL X(3)
      DIMENSION A(3),B(3)
C----------------------------------------------------------------------C
      CALL VECDIF (A,B,X)
      ALNGTH = AMAG(X)
      END

C**********************************************************************C
      FUNCTION TORSIN (A,B,C,D)
C----CALCULATE THE TORSION ANGLE FOR FOUR ATOMS
C     THE RESULT IS IN DEGREES
      DIMENSION X(3),Y(3),P(3),Q(3)
      DIMENSION A(3),B(3),C(3),D(3)
      DIMENSION ZERO (3)
      DATA ZERO /0.0, 0.0, 0.0/
C----------------------------------------------------------------------C
      CALL VECDIF(A,B,P)
      CALL VECDIF(C,B,Q)
      CALL CROSS(P,Q,X)
      CALL VECDIF(B,C,P)
      CALL VECDIF(D,C,Q)
      CALL CROSS(P,Q,Y)
      TORSIN = ANGLE (X,ZERO,Y)
      CALL CROSS (X,Y,Q)
      IF (DOT(P,Q) .GT. 0.0) TORSIN = -TORSIN
      END

C**********************************************************************C
      FUNCTION ANGLE (A,B,C)
C----GET THE ANGLE (IN DEGREES) BETWEEN TWO VECTORS DEFINED BY THE ENDS
C     OF THREE VECTORS
      DIMENSION X(3), Y(3)
      DIMENSION A(3),B(3),C(3)
C----------------------------------------------------------------------C
      CALL VECDIF (A,B,X)
      CALL VECDIF (C,B,Y)
      Q = AMAG(X) * AMAG(Y)
      IF (Q .LT. 1.0E-7) Q = 1.0E-7
      ANGLE = DOT (X,Y) / Q
      IF (ANGLE .GE. 1.0) THEN
         ANGLE = 0.0
      ELSEIF (ANGLE .LE. -1.0) THEN
         ANGLE = 180.0
      ELSE
         ANGLE = 57.2957795 * ACOS (ANGLE)
      ENDIF
      END

C**********************************************************************C
      SUBROUTINE INVER (A,B)
C----INVERT A THREE BY THREE MATRIX
      REAL X(9)
      DIMENSION A(9),B(9)
      DIMENSION IDATA1(3),IDATA2(3)
      DATA IDATA1 /4,7,1/
      DATA IDATA2 /7,1,4/
C----------------------------------------------------------------------C
      J = 0
      DO 1000 I=1,9,3
      J = J + 1
      ID1 = IDATA1(J)
      ID2 = IDATA2(J)
      CALL CROSS(A(ID1),A(ID2),X(I))
 1000 CONTINUE

      DETA = DET(A)
      IF (ABS(DETA) .LE. 1.0E-20) THEN
         DETA = SIGN(1.0E-20,DETA)
      ELSE 
         DETA = 1.0 / DETA
      ENDIF

      CALL CONMAT (X,DETA,X)
      CALL TRNSPZ(X,B)
      END

C**********************************************************************C
      SUBROUTINE MATMUL (A,B,C)
C----MULTIPLY TWO MATRICIES
      REAL X(9)
      DIMENSION A(9), B(9), C(9)
C----GET THE TRANSPOSE OF B INTO X
C----------------------------------------------------------------------C
      CALL TRNSPZ (B,X)
      IJ = 0
      DO 2000 I=1,9,3
      DO 2000 J=1,9,3
      IJ = IJ + 1
      C(IJ) = DOT(A(I),X(J))
 2000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE CART (CELL,AMAT)
C  FOR AN INPUT CELL 3 AXIAL EDGE LENGTHS AND THREE ANGLES (IN DEGREES
C  OR AS COSINES)
      REAL X(9)
      DIMENSION CELLD(3)
      DIMENSION CELL(6),AMAT(9)
C----------------------------------------------------------------------C
      CALL CPYVEC(CELL,CELLD)
      CALL CPYVEC(CELL(4),X)
      DO 1000 I=1,3
      IF (CELLD(I) .EQ. 0.0) CELLD(I) = 1.0
      IF (X(I) .LE. 1.1) X(I) = 57.2957795*ACOS(X(I))
      X(I) = X(I) / 57.2957795
 1000 CONTINUE
      SINX2=SIN(X(2))
      AMAT(1)=1.0
      AMAT(2)=0.0
      AMAT(3)=0.0
      AMAT(4)=COS(X(3))
      AMAT(5)=SIN(X(3))
      AMAT(6)=0.0
      AMAT(7)=COS(X(2))
      AMAT(8)=(COS(X(1))-AMAT(7)*AMAT(4))/AMAT(5)
      AMAT(9)=SQRT((SINX2-AMAT(8))*(SINX2+AMAT(8)))
      J=0
      DO 2000 I=1,9,3
      J=J+1
      CALL CONMUL (AMAT(I),CELLD(J),X(I))
 2000 CONTINUE
      CALL TRNSPZ (X,AMAT)
      END

C**********************************************************************C
      SUBROUTINE UNIT (X,Y)
C----RETURN A VECTOR Y WHICH IS THE DIRECTION COSINES OF X
      DIMENSION X(3),Y(3)
C----------------------------------------------------------------------C
      D = AMAG(X)
      IF (D - 1.0E-10) 9999, 9999, 1000
 1000 D = 1.0 / D
      CALL CONMUL (X,D,Y)
 9999 RETURN
      END

C**********************************************************************C
      SUBROUTINE CROSS (X,Y,Z)
C----COMPUTE Z = X CROSS Y
      DIMENSION X(3),Y(3),Z(3)
C----------------------------------------------------------------------C
      Z(1) = X(2)*Y(3) - Y(2)*X(3)
      Z(2) =-X(1)*Y(3) + Y(1)*X(3)
      Z(3) = X(1)*Y(2) - Y(1)*X(2)
      END

C**********************************************************************C
      FUNCTION AMAG (X)
C----GET AND RETURN THE MAGNITUDE OF A VECTOR X
      DIMENSION X(3)
C----------------------------------------------------------------------C
      AMAG = DOT(X,X)
      AMAG = SQRT(AMAG)
      END

C**********************************************************************C
      FUNCTION DOT (X,Y)
C----COMPUTE AND RETURN THE DOT PRODUCT OF X AND Y
      DIMENSION X(3),Y(3)
C----------------------------------------------------------------------C
      DOT = 0.0
      DO 1000 I=1,3
      DOT = DOT + X(I) * Y(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE VECSUM (X,Y,Z)
C----ADD TWO VECTORS AND RETURN THE SUM IN Z
      DIMENSION X(3), Y(3), Z(3)
C----------------------------------------------------------------------C
      DO 1000 I=1,3
      Z(I) = X(I) + Y(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE VECDIF (X,Y,Z)
C----SUBTRACT TWO VECTORS AND RETURN THE RESULT IN Z
      DIMENSION X(3), Y(3), Z(3)
C----------------------------------------------------------------------C
      DO 1000 I=1,3
      Z(I) = X(I) - Y(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE CONMUL (X,A,Y)
C----MULTIPLY A VECTOR TIMES A SCALAR
      DIMENSION X(3),Y(3)
C----------------------------------------------------------------------C
      DO 1000 I=1,3
      Y(I) = A * X(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE CONMAT(AMAT,X,BMAT)
      DIMENSION AMAT(9),BMAT(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,9
      BMAT(I) = X * AMAT(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE VEC (X,A,B,C)
C----COMPOSE A VECTOR X FROM THREE SCALARS
      DIMENSION X(3)
C----------------------------------------------------------------------C
      X(1) = A
      X(2) = B
      X(3) = C
      END

C**********************************************************************C
      SUBROUTINE TRNSPZ (A,B)
C----PUT THE TRANSPOSE OF A INTO B
      DIMENSION A(9), B(9)
C----------------------------------------------------------------------C
      J = 0
      DO 1000 I=1,9,3
      J = J + 1
      CALL UNVEC (A(I),B(J),B(J+3),B(J+6))
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE UNVEC (X,F,G,H)
C----RETURN THE VECTOR COMPONENTS AS SCALARS
      DIMENSION X(3)
C----------------------------------------------------------------------C
      F = X(1)
      G = X(2)
      H = X(3)
      END

C**********************************************************************C
      SUBROUTINE CPYMAT (A,B)
C----COPY MATRIX A INTO MATRIX B
      DIMENSION A(9), B(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,9,3
      CALL CPYVEC (A(I),B(I))
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE MAT (AMAT,A,B,C)
      DIMENSION AMAT(9)
C----------------------------------------------------------------------C
      CALL CPYVEC (A,AMAT(1))
      CALL CPYVEC (B,AMAT(4))
      CALL CPYVEC (C,AMAT(7))
      END

C**********************************************************************C
      SUBROUTINE UNMAT (A,X,Y,Z)
      DIMENSION A(9),X(3),Y(3),Z(3)
C----------------------------------------------------------------------C
      CALL CPYVEC (A(1),X)
      CALL CPYVEC (A(4),Y)
      CALL CPYVEC (A(7),Z)
      END

C**********************************************************************C
      SUBROUTINE CPYVEC (X,Y)
C----COPY A VECTOR X INTO A VECTOR Y
      DIMENSION X(3), Y(3)
C----------------------------------------------------------------------C
      DO 1000 I=1,3
      Y(I) = X(I)
 1000 CONTINUE
      END

C**********************************************************************C
      SUBROUTINE RANORI (ISEED,AMAT)
C
C derived from
C   real function randf( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns a rotation matrix for a uniformly distributed 
C        random rotation in 0-360 degrees about a 
C        uniformly distributed random vector
C
C    RANORI DOES NOT CALL RANVEC TO GET THE RANDOM VECTOR, SO THAT
C    THE USERS HAS THE OPTION OF GENERATING SEPARATE, DIFFERENT
C    STRINGS OF RANDOM ORIENTATIONS AND VECTORS
C

C  the algorithm is derived from one in J.M.Hammersley and D.C.
C  Handscomb, "Monte Carlo Methods," Methuen & Co., London and
C  Wiley & Sons, New York,1964, p47.  
C  The method is to generate a random unit quaterion, and
C  then to turn it into the corresponding rotation matrix.
C  Kearsley, Acta Cryst., v. A45, p. 208-210 (1989) solves a 
C  related linear least-squares problem.  The linearity
C  gives a hint that this method can be used.

      REAL    BUFFER(55)
      INTEGER INDX,JNDX,KNDX
      INTEGER ISEED,NC
      SAVE BUFFER, INDX, JNDX, KNDX
      REAL V(4), AMAT(9)
      DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GE. 0 .OR. INDX .LT. 0)
     2    CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)
 
      NC = 4
 1000 CONTINUE
      SUM = 0.0
      DO 2000 I=1,NC
         INDX = MOD(INDX ,55)+1
         JNDX = MOD(JNDX ,55)+1
         KNDX = MOD(KNDX ,55)+1
         BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
         V(I) = 2.0*BUFFER(INDX)-1.0
          SUM = SUM + V(I)**2
         IF (SUM .GT. 1.0) GO TO 1000
2000  CONTINUE
      SUM = SQRT(SUM)
      do 3000 I=1,4
         V(I) = V(I) /sum
 3000 continue
      AMAT(1) = V(1)**2 + V(2)**2 - V(3)**2 - V(4)**2
      AMAT(5) = V(1)**2 + V(3)**2 - V(2)**2 - V(4)**2
      AMAT(9) = V(1)**2 + V(4)**2 - V(2)**2 - V(3)**2

      AMAT(2) = 2.0*(V(2)*V(3) + V(1)*V(4))
      AMAT(3) = 2.0*(V(2)*V(4) - V(1)*V(3))
      AMAT(4) = 2.0*(V(2)*V(3) - V(1)*V(4))
      AMAT(6) = 2.0*(V(3)*V(4) + V(1)*V(2))
      AMAT(7) = 2.0*(V(2)*V(4) + V(1)*V(3))
      AMAT(8) = 2.0*(V(3)*V(4) - V(1)*V(2))
 
      END

C**********************************************************************C
      SUBROUTINE RANVEC (ISEED, V)
C
C derived from
C   real function randf( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns a uniformly distributed random vector with magnitude
C        in the range 0.0(+epsilon) to 1.0
C
      REAL    BUFFER(55)
      INTEGER INDX,JNDX,KNDX
      INTEGER ISEED,NC
      SAVE BUFFER, INDX, JNDX, KNDX
      REAL V(3)
      DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GE. 0 .OR. INDX .LT. 0)
     2   CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)
 
      NC = 3
 1000 CONTINUE
      DO 2000 I=1,NC
         INDX = MOD(INDX ,55)+1
         JNDX = MOD(JNDX ,55)+1
         KNDX = MOD(KNDX ,55)+1
         BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
         V(I) = 2.0*BUFFER(INDX)-1.0
2000  CONTINUE

C    REJECT THE VECTOR IF IT IS OUTSIDE OF THE UNIT SPHERE--
C    OTHERWISE THERE ARE TOO MANY RANDOM VECTORS IN THE CORNERS
C    OF THE SPHERE.  THAT IS, IT ISN'T GOOD ENOUGH TO GENERATE
C    THREE RANDOM NUMBERS AND THEN MAKE A UNIT VECTOR-- THEY
C    WOULD BE CONCENTRATED IN 8 SPOTS
C
C    HERE WE WANT VECTORS INSIDE OF A SPHERE
C
      IF (AMAG(V) .GT. 1.0) GO TO 1000
      END

C**********************************************************************C
      SUBROUTINE UNITRV (ISEED, V)
C
C derived from
C   real function randf( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns a uniformly distributed random vector 
C        with magnitude 1.0
C
      REAL    BUFFER(55)
      INTEGER INDX,JNDX,KNDX
      INTEGER ISEED,NC
      SAVE BUFFER, INDX, JNDX, KNDX
      REAL V(3)
      DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GE. 0 .OR. INDX .LT. 0)
     2   CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)
 
      NC = 3
 1000 CONTINUE
      DO 2000 I=1,NC
         INDX = MOD(INDX ,55)+1
         JNDX = MOD(JNDX ,55)+1
         KNDX = MOD(KNDX ,55)+1
         BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
         V(I) = 2.0*BUFFER(INDX)-1.0
2000  CONTINUE

C    REJECT THE VECTOR IF IT IS OUTSIDE OF THE UNIT SPHERE--
C    OTHERWISE THERE ARE TOO MANY RANDOM VECTORS IN THE CORNERS
C    OF THE SPHERE.  THAT IS, IT ISN'T GOOD ENOUGH TO GENERATE
C    THREE RANDOM NUMBERS AND THEN MAKE A UNIT VECTOR-- THEY
C    WOULD BE CONCENTRATED IN 8 SPOTS
C
C    HERE WE WANT VECTORS INSIDE OF A SPHERE
C
      IF (AMAG(V) .GT. 1.0) GO TO 1000
      CALL UNIT (V,V)
      END

C**********************************************************************C
      SUBROUTINE INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)
C   use a simple linear congruential random number generator to 
C   fill the buffer array to start up a linear shift register
C   generator
C   set the integers indx,jndx,kndx to the "magic" initial
C   values described by Knuth, The Art of Computer Programming,
C   vol 2, Seminumerical Algorithms

C   set iseed negative to indicate that the initialization was done
      REAL BUFFER(55)
      INTEGER ISEED, INDX, JNDX, KNDX
C----------------------------------------------------------------------C

      JNDX = ISEED
      DO 1000 INDX= 1,55
         JNDX = MOD(JNDX*2349 + 14867,32767)
         BUFFER(INDX) = ABS(FLOAT(JNDX)/32767)
 1000 CONTINUE
      ISEED = -ISEED
      IF (ISEED .EQ. 0) ISEED = -1
      INDX = 55
      KNDX = 54
      JNDX = 31
      END

C**********************************************************************C
      REAL FUNCTION RANDF( ISEED )
C   real function randf( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns a uniformly distributed random number in (0. < 1.)
C
      INTEGER ISEED
C
      INTEGER INDX,JNDX,KNDX
      REAL    BUFFER(55)
      SAVE BUFFER, INDX, JNDX, KNDX
      DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GE. 0 .OR. INDX .LT. 0)
     2   CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)

      INDX = MOD(INDX ,55)+1
      JNDX = MOD(JNDX ,55)+1
      KNDX = MOD(KNDX ,55)+1
      BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
      RANDF = BUFFER(INDX)
      END

C**********************************************************************C
      REAL FUNCTION RANDFG( ISEED )
C   real function randfg( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns a normally distributed random number with unit variance
C
C   EXACTLY FOLLOWS THE ALGORITHM OF KNUTH, THE ART OF COMPUTER
C   PROGRAMMING, V. 22D. ED., 1981, P. 117-118  (THE POLAR METHOD)
C   EXCEPT THAT LIMITS ARE INCLUDED BY MAKING THE MINIMUM VALUE OF "S"
C   BE EPS
C
      INTEGER ISEED
C
      INTEGER INDX,JNDX,KNDX
      REAL    BUFFER(55)
      REAL    EPS
      SAVE BUFFER, INDX, JNDX, KNDX,EPS
      DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GE. 0 .OR. INDX .LT. 0) THEN
         R1 = 1.
         DO 100 I=1,100
             EPS = R1
             R1 = R1/2
             IF( R1 + 1. .EQ. 1.) GOTO 101
100      CONTINUE
101      CONTINUE
         CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)
      ENDIF

 4000 CONTINUE
      INDX = MOD(INDX ,55)+1
      JNDX = MOD(JNDX ,55)+1
      KNDX = MOD(KNDX ,55)+1
      BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
      R1 = 2.0 * BUFFER(INDX) - 1.0
      INDX = MOD(INDX ,55)+1
      JNDX = MOD(JNDX ,55)+1
      KNDX = MOD(KNDX ,55)+1
      BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
      R2 = 2.0 * BUFFER(INDX) - 1.0
      S = R1*R1 + R2*R2
      IF (S .GE. 1.0) GO TO 4000
      IF (ABS(S) .LT. EPS) S = SIGN(EPS,S)
      RANDFG = R1 * SQRT(-2.0 * ALOG(S)/S)
      END

C**********************************************************************C
       SUBROUTINE BRAND( ISEED, X, NC )
C subroutine brand( iseed,x,nc)
C  generate nc random numbers in real array x
C
C derived from
C   real function randf( iseed )
C        if seed .gt. 0 use seed to initialize
C        returns NC uniformly distributed random numbers (-1. < 1.)
C
       INTEGER ISEED,NC
       REAL X(NC)
C
       INTEGER INDX,JNDX,KNDX
       REAL    BUFFER(55)
       SAVE BUFFER,INDX,JNDX,KNDX
       DATA INDX /-1/
C----------------------------------------------------------------------C

      IF( ISEED .GT. 0 .OR. INDX .LT. 0)
     2     CALL INITRN (ISEED,INDX,JNDX,KNDX,BUFFER)

      DO 1000 I=1,NC
         INDX = MOD(INDX ,55)+1
         JNDX = MOD(JNDX ,55)+1
         KNDX = MOD(KNDX ,55)+1
         BUFFER(INDX) = AMOD( BUFFER(JNDX) + BUFFER(KNDX), 1.0)
         X(I) = 2.0*BUFFER(INDX)-1.0
1000  CONTINUE
      END

C**********************************************************************C
      SUBROUTINE BUILDV (X1,X2,X3,ALNG,ANGL,TRSN,X4)

C    SUBROUTINE TO RETURN THE COORDINATES OF A GENERATED ATOM (X4)
C
C    INPUT PARAMETERS ARE THE COORDINATES OF THREE ATOMS (X1,X2,X3)
C    AND THE LENGTH OF THE NEW BOND (ALNG) WHICH IS ATTACHED TO X3,
C    THE ANGLE OF THE NEW BOND WITH ATOMS X2 AND X3, AND THE NEW
C    TORSION ANGLE
C
C    FOR THE USERS' CONVENIENCE, BUILDV HAS BEEN SET UP SO THAT THE
C    1) THE INITIAL CALL CAN HAVE X1,X2,X3 IDENTICAL
C    2) THE SECOND CALL CAN HAVE X1,X2 IDENTICAL
C    THIS MEANS THAT NO SPECIAL "STARTUP" CODE HAS TO BE WRITTEN

C      ALGORITHM

C    THE ATOMS ARE TRANSLATED SO THAT THE MIDDLE ATOM OF THE INPUT
C    THREE IS AT THE ORIGIN
C
C    PAIR THEN GIVES THE MATRIX (AMAT) WHICH WILL ROTATE THE
C    TRANSLATED ATOMS SO THAT X3 (NOW V3) IS ON THE X-AXIS AND
C    X1 (NOW V1) IS IN THE THIRD QUADRANT
C
C    THE COORDINATES OF THE NEW ATOM ( V ) ARE THEN CALCULATED IN
C    THIS ROTATED FRAME OF REFERENCE, THE MATRIX IS INVERTED (IT
C    WOULD BE SUFFICIENT TO JUST TRANSPOSE), THE NEW ATOM-S
C    COORDINATES ARE TRANSFORMED BACK TO THE ORIGINAL REFERENCE 
C    SYSTEM, AND THEN TRANSLATED BACK BY THE COMPONENTS OF X3

      DIMENSION X1(3),X2(3),X3(3),X4(3)
      DIMENSION V(3),V1(3),V3(3),AMAT(9),BMAT(9),XAXIS(3),DIAG(3)
      DATA XAXIS /1.,0.0,0.0/
      DATA DIAG /-1.18734,-3.14185,0.0/
      CALL VECDIF (X1,X2,V1)
      CALL VECDIF (X3,X2,V3)
      CALL PAIR (V3,V1,XAXIS,DIAG,AMAT)
      COST = COS(TRSN)
      SINA = SIN(ANGL)
      V(1) =-ALNG * COS(ANGL)
      V(2) = -ALNG * SINA * COST
      V(3) =-ALNG * SINA * SIN(TRSN)
      CALL INVER (AMAT,BMAT)
      CALL VECMUL (V,BMAT,X4)
      CALL VECSUM (X4,X3,X4)
      RETURN
      END


C**********************************************************************C
      SUBROUTINE GETCM (N,X,CM)
      REAL X(3,N),CM(3)
      CALL VEC(CM,0.0,0.0,0.0)
      IF (N .GT. 0) THEN
         DO 1000 I=1,N
         CALL VECSUM (CM,X(1,I),CM)
 1000    CONTINUE
         CALL CONMUL (CM,1.0/FLOAT(N),CM)
      ELSE
         CALL VEC (CM, 0.0,0.0,0.0)
      ENDIF
      END


C**********************************************************************C
      SUBROUTINE BSTPLN (NPTS,X,CM,V)
      REAL X(3,NPTS),CM(3),V(3)
      REAL AMAT(9),BMAT(9)
C----------------------------------------------------------------------C

      CALL VEC (CM,0.0,0.0,0.0)
      CALL VEC (V ,0.0,0.0,0.0)
      IF (NPTS .LE. 0) RETURN

      CALL GETCM (NPTS,X,CM)

      DO 2000 I=1,NPTS
         CALL VECDIF (X(1,I),CM,V)
         CALL INRTIA(V,1.0,AMAT)
 2000 CONTINUE

      E1 = EIGEN1(AMAT,V,BMAT)
      END

C**********************************************************************C
      SUBROUTINE BSTLIN (NPTS,X,CM,V)
      REAL X(3,NPTS),CM(3),V(3)
      REAL AMAT(9),BMAT(9)
C----------------------------------------------------------------------C

      CALL VEC (CM,0.0,0.0,0.0)
      CALL VEC (V ,0.0,0.0,0.0)
      IF (NPTS .LE. 0) RETURN

      CALL GETCM (NPTS,X,CM)

      DO 2000 I=1,NPTS
         CALL VECDIF (X(1,I),CM,V)
         CALL INRTIA(V,1.0,AMAT)
 2000 CONTINUE
      CALL INVER(AMAT,BMAT)
      E1 = EIGEN1(BMAT,V,AMAT)
      END


C**********************************************************************C
      SUBROUTINE BSTROT (NATOMS,ATM1,ATM2,DELCM,ERROR,B,TEST)

C  The algorithm is from Ferro and Hermans, Acta Cryst., A33,345-347
C  (1977).  Other algorithms give the same result, such as:
C
C  Kearsley, Acta Cryst., A45, 208-210 (1989)
C  Diamond,  Acta Cryst., A44, 211-216 (1988)
C  Kabsch,   Acta Cryst., A34, 827-828 (1978)
C  Kabsch,   Acta Cryst., A32, 922-923 (1976)
C
C  The algorithm of Ferro and Hermans has the advantage that the 
C  rotational components are removed iteratively, so that residual
C  numerical errors are correctly removed.  This leaves no
C  possibility of poorly orthogonalized matrices.  The explicit
C  iteration is not a problem.  The other methods simply hide the
C  iteration in the eigenvector solver.  Ferro and Hermans
C  simply build what is a simple, special purpose version of the QR
C  method.

      CHARACTER *6 TEST
      REAL ATM1(3,NATOMS), ATM2(3,NATOMS)
      REAL CM1(3),CM2(3),DELCM(3)
      REAL V(3), V1(3), V2(3)
      REAL AM(9)
      REAL XMAT(9),YMAT(9),ZMAT(9)
      REAL A(9),B(9),C(9),D(9),R(9)
C----------------------------------------------------------------------C

      IF (TEST .NE. 'BSTROT' .AND. TEST .NE. 'bstrot')
     2 CALL BADCAL (TEST,'BSTROT')

      IF (NATOMS .LE. 0) THEN
         CALL VEC (DELCM, 0.0,0.0,0.0)
         CALL UNITMT (B)
         ERROR = 0.0
         RETURN
      ENDIF

      CALL GETCM (NATOMS,ATM1,CM1)
      CALL GETCM (NATOMS,ATM2,CM2)
      CALL VECDIF (CM2,CM1,DELCM)

      E0 = 0
      CALL ZERMAT(R)
      DO 2000 I=1,NATOMS
         CALL VECDIF (ATM1(1,I),CM1,V1)
         CALL VECDIF (ATM2(1,I),CM2,V2)
         CALL TENPRD(V2,V1,AM)
         CALL MATSUM (R,AM,R)
 2000 CONTINUE

      CALL CPYMAT (R,A)
      CALL UNITMT(R)
      DO 3000 I=1,20
         CALL UNX(A,XMAT)
         CALL MATMUL (XMAT,A,B)
         CALL UNY(B,YMAT)
         CALL MATMUL (YMAT,B,C)
         CALL UNZ(C,ZMAT)
         CALL MATMUL (ZMAT,C,B)
         CALL CPYMAT (B,A)

         CALL MATMUL (XMAT,R,D)
         CALL MATMUL (YMAT,D,C)
         CALL MATMUL (ZMAT,C,R)
         IF (ABS(XMAT(7))+ABS(YMAT(3))+ABS(ZMAT(3)) .LT. 1.0E-8)
     2       GO TO 4000
 3000    CONTINUE
 4000 CONTINUE
      CALL TRNSPZ (R,B)

      ERROR = 0.0
      DO 5000 I=1,NATOMS
         CALL VECDIF (ATM1(1,I),CM1,V1)
         CALL VECDIF (ATM2(1,I),CM2,V2)
         CALL VECMUL (V1,B,V)
         CALL VECDIF (V,V2,V)
         ERROR = ERROR + DOT(V,V)
 5000 CONTINUE
      ERROR = SQRT(ERROR)
      END

C**********************************************************************C
      SUBROUTINE UNX (A,B)
      REAL A(9),B(9)
      REAL XAXIS(3)
      DATA XAXIS /1.0,0.0,0.0/
C----------------------------------------------------------------------C
      ANG = ATAN2(A(8)-A(6),A(5)+A(9))
      CALL ROTMAT (XAXIS,-ANG,B)
      END

C**********************************************************************C
      SUBROUTINE UNY (A,B)
      REAL A(9),B(9)
      REAL YAXIS(3)
      DATA YAXIS /0.0,1.0,0.0/
C----------------------------------------------------------------------C
      ANG = ATAN2(A(7)-A(3),A(1)+A(9))
      CALL ROTMAT (YAXIS,ANG,B)
      END

C**********************************************************************C
      SUBROUTINE UNZ (A,B)
      REAL A(9),B(9)
      REAL ZAXIS(3)
      DATA ZAXIS /0.0,0.0,1.0/
C----------------------------------------------------------------------C
      ANG = ATAN2(A(4)-A(2),A(1)+A(5))
      CALL ROTMAT (ZAXIS,-ANG,B)
      END

C**********************************************************************C
      SUBROUTINE ZERMAT (AM)
      REAL AM(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,9
 1000 AM(I) = 0.0
      END

C**********************************************************************C
      SUBROUTINE UNITMT(AM)
      REAL AM(9)
C----------------------------------------------------------------------C
      CALL ZERMAT(AM)
      DO 1000 I=1,9,4
 1000 AM(I) = 1.0
      END

C**********************************************************************C
      SUBROUTINE EIGEN3 (AM, EVAL, EV1,EV2,EV3)
      REAL AM(9), EVAL(3), EV1(3), EV2(3), EV3(3)
      REAL TM(9), AMI(9)
C----------------------------------------------------------------------C

      DUMMY = EIGEN1 (AM, EV1, TM)
      CALL INVER (AM,AMI)
      DUMMY = EIGEN1 (AMI, EV3, TM)

      CALL CROSS (EV3, EV1, EV2)

      CALL VECMUL (EV1,AM,V)
      EVAL(1) = SQRT(AMAG(V))
      CALL VECMUL (EV2,AM,V)
      EVAL(2) = SQRT(AMAG(V))
      CALL VECMUL (EV3,AM,V)
      EVAL(3) = SQRT(AMAG(V))
      END
    
C**********************************************************************C
      SUBROUTINE MATSUM (A,B,C)
      REAL A(9),B(9),C(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,9
 1000 C(I) = A(I) + B(I)
      END

C**********************************************************************C
      SUBROUTINE MATDIF (A,B,C)
      REAL A(9),B(9),C(9)
C----------------------------------------------------------------------C
      DO 1000 I=1,9
 1000 C(I) = A(I) - B(I)
      END

C**********************************************************************C
      SUBROUTINE TENPRD (X,Y,AM)
      REAL X(3), Y(3), AM(9)
C----------------------------------------------------------------------C
      CALL CONMUL (Y(1),X(1),AM(1))
      CALL CONMUL (Y(1),X(2),AM(4))
      CALL CONMUL (Y(1),X(3),AM(7))
      END      

C**********************************************************************C
      SUBROUTINE BADCAL (A,B)
      CHARACTER *6 A,B
C----------------------------------------------------------------------C
      WRITE (*,*)  ' BAD SUBROUTINE CALL TO ',B
      WRITE (*,*)  ' CALLING NAME =',A
      STOP
      END

C**********************************************************************C
      SUBROUTINE BLDTRE (MXTREE,X,IDIN,TREE,TEST)
      PARAMETER (NVEC=3)
      CHARACTER*6 TEST
      LOGICAL DEBUG
      INTEGER RMAX
      REAL TREE(MXTREE)
      DATA LINK,RMAX,ID,ICHILD  /1,2,3,4/
      DATA DEBUG /.FALSE./
C----------------------------------------------------------------------C
 
      IF (TEST .NE. 'BLDTRE' .AND. TEST .NE. 'bldtre')
     2     CALL BADCAL (TEST,'BLDTRE')
 
C   THE NODE SIZE IS 4 PLUS THE SIZE OF THE VECTOR
      NODSIZ = 4+NVEC
      IPOINT = 2
      IF (TREE(1) .GT. 0) THEN
          IFREE = TREE(1)
      ELSE 
         IFREE = 2
      ENDIF
 1000 CONTINUE
      IF (DEBUG) WRITE (*,*)
     2 ' AFTER 1000 IN BLDTRE, IPOINT,TREE(IPOINT) ',
     3   IPOINT,TREE(IPOINT)
      IF (DEBUG) WRITE (*,*) 'IFREE,TREE(1),TREE(2) ',
     2     IFREE,INT(TREE(1)),INT(TREE(2))
 
      IF (TREE(IPOINT) .EQ. 0) THEN
         IF (DEBUG) WRITE (*,*) ' A NEW NODE IS BEING ALLOCATED'
         IPOINT = IFREE
         TREE(IPOINT) = -1
         TREE(IPOINT+LINK) = -1
         TREE(IPOINT+ID) = IDIN
         CALL CPYVEC (X,TREE(IPOINT+ICHILD))
         TREE(1) = IFREE + NODSIZ
         RETURN
      ELSEIF (TREE(IPOINT) .EQ. -1) THEN
         IF (DEBUG) WRITE (*,*) ' RIGHT CHILD OF NODE IS BEING FILLED'
         TREE(IPOINT) = IFREE
         TREE(IFREE+LINK) = -1
         TREE (IFREE+RMAX) = -1.0
         TREE(IFREE+ID) = IDIN
         CALL CPYVEC (X,TREE(IFREE+ICHILD))
         IFREE = IFREE + NODSIZ
         TREE(1) = IFREE
         RETURN
      ELSE
         IRIGHT = TREE(IPOINT)
         DL = ALNGTH (X,TREE(IPOINT+ICHILD))
         DR = ALNGTH (X,TREE(IRIGHT+ICHILD))
         IF (DEBUG) WRITE (*,*) ' DL,DR ',DL,DR
         IF (DR .GT. DL) THEN
             IF (DEBUG) WRITE (*,*) DR,DL,IPOINT,LINK,RMAX
             IF (TREE(IPOINT+LINK) .LE. 0) THEN                
                TREE(IPOINT+RMAX) = DL
                TREE(IPOINT+LINK) = IFREE
                IPOINT = IFREE
             ELSE 
                TREE(IPOINT+RMAX) = 
     2                 MAX(DL,TREE(IPOINT+RMAX))
                IPOINT = TREE(IPOINT+LINK)
             ENDIF
 
         ELSE
             IF (TREE(IRIGHT+LINK) .LE. 0) THEN
                 TREE(IRIGHT+RMAX) = DR
                 TREE(IRIGHT+LINK) = IFREE
                 IPOINT = IFREE
              ELSE
                 TREE(IRIGHT+RMAX) = 
     2                  MAX(DR,TREE(IRIGHT+RMAX))
                 IPOINT = TREE(IRIGHT+LINK)
              ENDIF
 
         ENDIF
         GO TO 1000
      ENDIF
      END
 
C**********************************************************************C
      FUNCTION NEARST (MXTREE,X,RADMAX,TREE,IP,IDOUT,TEST)
      CHARACTER*6 TEST
      LOGICAL DEBUG
      INTEGER ISTAK(1000)
      REAL TREE(MXTREE)
      INTEGER RMAX
      DATA DEBUG /.FALSE./
      DATA LINK,RMAX,ID,ICHILD /1,2,3,4/
      DATA RIGHT,LEFT,END /111,112,113/
C----------------------------------------------------------------------C

      IF (TEST .NE. 'NEARST' .AND. TEST .NE. 'nearst')
     2    CALL BADCAL (TEST,'NEARST') 

      ISTKP = 0
      IP = 0
      IPOINT = 2
      CURMIN = RADMAX
      DIR = LEFT
      DIRPRV = RIGHT
      IPREV = IPOINT -1
 1000 CONTINUE
      IPREV = IPOINT
      DIRPRV = DIR
      IF (DEBUG) WRITE (*,*)  ' IN NEARST 1000, IPOINT = ',IPOINT
      IF (TREE(IPOINT) .EQ. 0) THEN
         IF (DEBUG) WRITE (*,*) ' AT AN END WITH IPOINT = ',IPOINT
         DIR = END
      ELSEIF (DIR .EQ. RIGHT) THEN
         IRIGHT = TREE(IPOINT)
         IF (DEBUG) WRITE (*,*) ' WENT RIGHT WITH IPOINT ', IPOINT
         DR = ALNGTH (TREE(IRIGHT+ICHILD),X)
         IF (DR .LT. CURMIN) THEN
            CURMIN = DR
            IP = IRIGHT + ICHILD
            IDOUT = TREE(IRIGHT+ID)
         ENDIF
         IF (TREE(IRIGHT+LINK) .LE. 0) THEN
            IF (DEBUG) WRITE (*,*) ' ON RIGHT BRANCH, UNSTACK A POINT'
            DIR = END
         ELSEIF (TREE(IRIGHT+RMAX)+CURMIN .GT. DR) THEN
            IPOINT = TREE(IRIGHT+LINK)
            DIR = LEFT      
         ELSE
             DIR = END
         ENDIF
      ELSE
         IF (DEBUG) WRITE (*,*) ' WENT LEFT, IPOINT ',IPOINT
         DIR = LEFT
         IF (TREE(IPOINT) .gt. 0) THEN
            IF (DEBUG) WRITE (*,*) ' STACK ONE '
            CALL STACK(IPOINT,ISTAK,ISTKP)
         ENDIF
         DL = ALNGTH(TREE(IPOINT+ICHILD),X)
         IF (DL .LT. CURMIN) THEN
            CURMIN = DL
            IP = IPOINT+ICHILD
            IDOUT = TREE(IPOINT+ID)
         ENDIF
         IF (TREE(IPOINT+LINK) .LE. 0) THEN
            IF (DEBUG) WRITE (*,*) ' NO LEFT LINK, GO BACK'
            DIR = END
         ELSEIF (TREE(IPOINT+RMAX) .LT. 0.0) THEN
            IF (DEBUG) WRITE (*,*) ' NO DESCENDING LEFT TREE'
            DIR = END
         ELSEIF (TREE(IPOINT+RMAX)+CURMIN .GT. DL) THEN
            IF (DEBUG) WRITE (*,*) ' GOING TO GO DOWN ONE LEVEL ',
     2         ' IPOINT AND UPDATE ',IPOINT,' ',INT(TREE(IPOINT+LINK))
            IPOINT = TREE(IPOINT+LINK)
         ELSE
           IF (DEBUG) WRITE (*,*) ' NO CLOSER POINTS ON LEFT '
           IF (DEBUG) WRITE (*,*) 'CURMIN,TREE(IPOINT+RMAX),DL ',
     2        CURMIN,TREE(IPOINT+RMAX),DL
           DIR = END
         ENDIF
      ENDIF
      IF (DIR .EQ. END) THEN
         IF (IUNSTK(IPOINT,ISTAK,ISTKP) .LE. 0) GO TO 8000
         DIR = RIGHT
      ENDIF
      GO TO 1000
 8000 CONTINUE
      NEARST = IP
      END
 
C**********************************************************************C
      SUBROUTINE STACK(NEXT,ISTAK,ISTKP)
      INTEGER ISTAK(1000)
      LOGICAL DEBUG
      DATA DEBUG /.FALSE./
C----------------------------------------------------------------------C
      IF (DEBUG) WRITE (*,*) ' IN STACK, STACK POINTER,IPOINT ',
     2    ISTKP,NEXT
      ISTKP = ISTKP + 1
      ISTAK(ISTKP) = NEXT
      END
 
C**********************************************************************C
      FUNCTION IUNSTK(NEXT,ISTAK,ISTKP)
      INTEGER ISTAK(1000)
      LOGICAL DEBUG
      DATA DEBUG /.FALSE./
C----------------------------------------------------------------------C
      IF (DEBUG) WRITE (*,*) ' IUNSTK,NEXT ',NEXT
      IF (ISTKP .GT. 0) THEN
         NEXT = ISTAK(ISTKP)
         ISTKP = ISTKP-1
      ELSE
         NEXT = 0
      ENDIF
      IUNSTK = NEXT
      END
 
C**********************************************************************C
      FUNCTION INSPHR (MXTREE,X,RADMAX,TREE,
     2    MXLIST,NLIST,LIST,IDLIST,TEST)
      CHARACTER*6 TEST
      LOGICAL DEBUG
      INTEGER ISTAK(1000)
      REAL TREE(MXTREE)
      INTEGER LIST(MXLIST),IDLIST(MXLIST)
      INTEGER RMAX
      DATA DEBUG /.FALSE./
      DATA LINK,RMAX,ID,ICHILD /1,2,3,4/
      DATA RIGHT,LEFT,END /111,112,113/
C----------------------------------------------------------------------C
 
      IF (TEST .NE. 'INSPHR' .AND. TEST .NE. 'insphr')
     2    CALL BADCAL (TEST,'INSPHR')
 
      ISTKP = 0
      NLIST = 0
      IPOINT = 2
      CURMIN = RADMAX
      DIR = LEFT
      DIRPRV = RIGHT
      IPREV = IPOINT -1
 1000 CONTINUE
      IF (IPREV .EQ. IPOINT .AND. DIRPRV .EQ. DIR) THEN
         WRITE (*,*) ' INTERNAL ERROR IN INSPHR '
         WRITE (*,*) ' TREE POINTER DIDN''T CHANGE',IPOINT,' ',DIR
         STOP
      ELSEIF (IPOINT .EQ. 0) THEN
         WRITE (*,*) ' INTERNAL ERROR IN INSPHR, IPOINT = 0'
         STOP
      ENDIF
      IPREV = IPOINT
      DIRPRV = DIR
      IF (DEBUG) WRITE (*,*)  ' IN INSPHR 1000, IPOINT = ',IPOINT
      IF (TREE(IPOINT) .EQ. 0) THEN
         IF (DEBUG) WRITE (*,*) ' AT AN END WITH IPOINT = ',IPOINT
         DIR = END
      ELSEIF (DIR .EQ. RIGHT) THEN
         IRIGHT = TREE(IPOINT)
         IF (DEBUG) WRITE (*,*) ' WENT RIGHT WITH IPOINT ', IPOINT
         DR = ALNGTH (TREE(IRIGHT+ICHILD),X)
         IF (DR .LT. CURMIN) THEN
            NLIST = NLIST + 1
            IF (NLIST .GT. MXLIST) GO TO 8000
            LIST(NLIST) = IRIGHT + ICHILD
            IDLIST(NLIST) = TREE(IRIGHT+ID)
         ENDIF
         IF (TREE(IRIGHT+LINK) .LE. 0) THEN
            IF (DEBUG) WRITE (*,*) ' ON RIGHT BRANCH, UNSTACK A POINT'
            DIR = END
         ELSEIF (TREE(IRIGHT+RMAX)+CURMIN .GT.
     2          ALNGTH(TREE(IRIGHT+ICHILD),X) ) THEN
            IPOINT = TREE(IRIGHT+LINK)
            DIR = LEFT      
         ELSE
             DIR = END
         ENDIF
      ELSE
         IF (DEBUG) WRITE (*,*) ' WENT LEFT, IPOINT ',IPOINT
         DIR = LEFT
         IF (TREE(IPOINT) .gt. 0) THEN
            IF (DEBUG) WRITE (*,*) ' STACK ONE '
            CALL STACK(IPOINT,ISTAK,ISTKP)
         ENDIF
         DL = ALNGTH(TREE(IPOINT+ICHILD),X)
         IF (DL .LT. CURMIN) THEN
            NLIST = NLIST + 1
            IF (NLIST .GT. MXLIST) GO TO 8000
            LIST(NLIST) = IPOINT+ICHILD
            IDLIST(NLIST) = TREE(IPOINT+ID)
         ENDIF
         IF (TREE(IPOINT+LINK) .LE. 0) THEN
            IF (DEBUG) WRITE (*,*) ' NO LEFT LINK, GO BACK'
            DIR = END
         ELSEIF (TREE(IPOINT+RMAX) .LT. 0.0) THEN
            IF (DEBUG) WRITE (*,*) ' NO DESCENDING LEFT TREE'
            DIR = END
         ELSEIF (TREE(IPOINT+RMAX)+CURMIN .GT. DL) THEN
            IF (DEBUG) WRITE (*,*) ' GOING TO GO DOWN ONE LEVEL ',
     2         ' IPOINT AND UPDATE ',IPOINT,' ',INT(TREE(IPOINT+LINK))
            IPOINT = TREE(IPOINT+LINK)
         ELSE
           IF (DEBUG) WRITE (*,*) ' NO CLOSER POINTS ON LEFT '
           IF (DEBUG) WRITE (*,*) 'CURMIN,TREE(IPOINT+RMAX),DL ',
     2        CURMIN,TREE(IPOINT+RMAX),DL
           DIR = END
         ENDIF
      ENDIF
      IF (DIR .EQ. END) THEN
         IF (IUNSTK(IPOINT,ISTAK,ISTKP) .LE. 0) GO TO 8000
         DIR = RIGHT
      ENDIF
      GO TO 1000
 8000 CONTINUE
      IF (NLIST .LE. MXLIST) THEN
         INSPHR = NLIST
      ELSE
         INSPHR = -NLIST
         NLIST = MXLIST
      ENDIF
      END

C**********************************************************************C
      FUNCTION ARCOS(X)
C----------------------------------------------------------------------C
      C2 = X*X
      C = X
      IF  (C2 .LE. 1.0) GO TO 2000
      IF (ABS(X) .LT. 1.00001) GO TO 1000
      ARCOS = 0.0
      RETURN
 1000 C2 = 1.0
      C = SIGN(1.0,X)
 2000 S = SQRT(1.0 - C2)
      ARCOS = ATAN2 (S,C)
      RETURN
      END

C**********************************************************************C
      FUNCTION ACOS(X)
C----------------------------------------------------------------------C
      C2 = X*X
      C = X
      IF  (C2 .LE. 1.0) GO TO 2000
      IF (ABS(X) .LT. 1.00001) GO TO 1000
      ACOS = 0.0
      RETURN
 1000 C2 = 1.0
      C = SIGN(1.0,X)
 2000 S = SQRT(1.0 - C2)
      ACOS = ATAN2 (S,C)
      RETURN
      END

C**********************************************************************C
      FUNCTION ARSIN(X)
C----------------------------------------------------------------------C
      S2 = X*X
      S = X
      IF (S2 .LT. 1.0) GO TO 2000
      IF (ABS(X) .LT. 1.00001) GO TO 1000
      ARSIN = 0.0
      RETURN
 1000 S2 = 1.0
      S = SIGN(1.0,X)
 2000 C = SQRT(1.0-S2)
      ARSIN = ATAN2(S,C)
      RETURN
      END

C**********************************************************************C
      FUNCTION ASIN(X)
C----------------------------------------------------------------------C
      S2 = X*X
      S = X
      IF (S2 .LT. 1.0) GO TO 2000
      IF (ABS(X) .LT. 1.00001) GO TO 1000
      ASIN = 0.0
      RETURN
 1000 S2 = 1.0
      S = SIGN(1.0,X)
 2000 C = SQRT(1.0-S2)
      ASIN = ATAN2(S,C)
      RETURN
      END
